#!/usr/bin/ruby
#
# output .dot file of function graph
#

require 'ftable'
require 'getopts'


def main
  getopts('nh', 'r:', 'help') or usage 1
  usage 0 if $OPT_help
  usage 0 if $OPT_h
  
  table = FunctionTable.parse_files(ARGV)
  reject_useless_nodes table, $OPT_r
  exit 0 if $OPT_n
  output_dot_file table
end

def usage( status )
  (status == 0 ? $stdout : $stderr).print <<EOS
Usage: #{File.basename $0} <source file> [<reject file>]
EOS
  exit status
end

def output_dot_file( table )
  puts 'digraph G {'
  puts '    node [shape=plaintext];'
  puts
  table.each_key do |name|
    puts %Q(    #{name} [label="#{name}"];)
  end
  puts
  table.each do |name, data|
    data.refs.each do |dest|
      puts %Q(    #{name} -> #{dest};)
    end
  end
  puts '}'
end


def reject_useless_nodes( table, rejfile )
  return table unless rejfile

  curr = nil
  oplineno = nil
  File.open(rejfile) {|f|
      f.each do |line|
        line = line.split(/#/, 2).first
        case line
        when /\A\w/
          eval_command table, curr, oplineno if curr
          curr = line.strip
          oplineno = f.lineno
        else
          curr << ' ' << line.strip if curr
        end
      end
      eval_command table, curr, oplineno if curr
  }
end

def eval_command( table, line, lineno )
  op, arg = line.strip.split(/\s+/, 2)
  case op.gsub(/_/, '').downcase
  when 'mark'
    arg.split.each do |n|
      table.mark n
    end
  when 'markrecursive'
    table.mark_recursive arg
  when 'markaround'
    table.mark_around arg
  when 'rejectunmarked'
    table.reject_unmarked

  when 'rejecteq', 'del', 'delete', 'remove', 'rm'
    re = words_pattern(arg.split)
    table.reject! {|name, m| re === name }
  when 'rejectmatch'
    re = unified_pattern(arg.split)
    table.reject! {|name, m| re === name }
  when 'rejectnotmatch'
    re = unified_pattern(arg.split)
    table.reject! {|name, m| not re === name }

  when 'rejectexternal'
    table.remove_links_to_external_function
  when 'rejectextern'
    table.remove_links_to_extern_function
  when 'removeselflink'
    table.remove_self_link

  when 'rejectdags'
    i = arg.to_i
    table.reject_DAGs (i == 0 ? nil : i)
  when 'rejectisolated'
    table.reject_isolated_functions
  else
    $stderr.puts "no such operation: #{op} (at line #{lineno})"
    exit 1
  end
end

def unified_pattern( patterns )
  Regexp.compile('(?:' + patterns.join('|') + ')')
end

def words_pattern( list )
  Regexp.compile('\A(?:' +
      list.map {|w| Regexp.escape(w) }.join('|') \
  + ')\z')
end


main
